/** @file
  Libray implementation for Flash Device Library based on Multiple Flash Support
  library intance.

Copyright(c) 2013 Intel Corporation. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

* Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in
the documentation and/or other materials provided with the
distribution.
* Neither the name of Intel Corporation nor the names of its
contributors may be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

**/
#include <Uefi/UefiBaseType.h>
#include <Uefi/UefiSpec.h>
#include <Library/DebugLib.h>
#include <Library/BaseMemoryLib.h>
#include <Library/PcdLib.h>
#include <Library/QuarkBootRomLib.h>

/**
  Security library to authenticate the Firmware Volume

  @param ImageBaseAddress         Pointer to the current Firmware Volume under
                                  consideration

  @retval EFI_SUCCESS             Firmware Volume is legal
  @retval EFI_SECURITY_VIOLATION  Firmware Volume is NOT legal.

**/
EFI_STATUS
EFIAPI
SecurityAuthenticateImage (
  IN VOID *ImageBaseAddress
  )
{
  BOOLEAN                 (*ValidateModuleCallBack)(QuarkSecurityHeader_t*, RSA2048PublicKey_t*, ScratchMemory_t*) = NULL;  // bootrom code validate entry function pointer
  UINT8                   MemoryBufferArray[QUARK_CRYPTO_HEAP_SIZE_BYTES];        // Allocate a heap to be used by the BootRom crypto functions
  UINT32                  *ValidateModuleCallBackAddress = 0x0;
  ScratchMemory_t         *scratchAreaInfo;
  RSA2048PublicKey_t      *OemRsaKeyAddress;
  QuarkSecurityHeader_t   *SignedImageBaseAddress;  // Location in memory of image to be validated
  EFI_STATUS              Status;
  BOOLEAN                 RomResult;

  //
  // Get the address of the OEM RSA Public Key
  //
  SignedImageBaseAddress = (QuarkSecurityHeader_t *)QUARK_KEY_MODULE_BASE_ADDRESS;
  OemRsaKeyAddress = (RSA2048PublicKey_t *)(((UINT32)SignedImageBaseAddress->ModuleHeaderSizeBytes) + (UINT32)QUARK_KEY_MODULE_BASE_ADDRESS);

  //
  // Get the address of the image to be authenticated (including the CSH)
  //
  SignedImageBaseAddress = (QuarkSecurityHeader_t *)((UINT32)ImageBaseAddress - PcdGet32 (PcdFvSecurityHeaderSize));

  //
  // Initialise memory to pass as heap to Validate function to all zeros
  //
  ZeroMem (MemoryBufferArray, QUARK_CRYPTO_HEAP_SIZE_BYTES);

  //
  // Init the Scratch Memory Structure.
  //
  scratchAreaInfo = (ScratchMemory_t *)&MemoryBufferArray;
  scratchAreaInfo->HeapStart = (UINT8 *)(scratchAreaInfo + 1);                         // Next address after the structure itself - use pointer++ to figure that address for us.
  scratchAreaInfo->HeapEnd = (UINT8 *)(scratchAreaInfo) + QUARK_CRYPTO_HEAP_SIZE_BYTES;
  scratchAreaInfo->NextFreeMem = scratchAreaInfo->HeapStart;                           // Reset our next free memory pointer to the base address.
  scratchAreaInfo->FatalCode = 0;
  scratchAreaInfo->DebugCode = 0;

  //
  // Validate Key Module using Oem Key.
  //
  ValidateModuleCallBackAddress = (UINT32*)QUARK_BOOTROM_VALIDATE_FUNCTION_ENTRYPOINT_ADDRESS;
  ValidateModuleCallBack =  (BOOLEAN(*)(QuarkSecurityHeader_t*, RSA2048PublicKey_t*, ScratchMemory_t*))((UINT32)(*ValidateModuleCallBackAddress)); // actual address here to jump to make sure to get it each time!

  RomResult = ValidateModuleCallBack (
                (QuarkSecurityHeader_t*) SignedImageBaseAddress,
                (RSA2048PublicKey_t*) OemRsaKeyAddress,
                (ScratchMemory_t*) scratchAreaInfo
                );

  DEBUG ((EFI_D_INFO, "SecurityAuthenticateImage RomResult:%d DebugCode:0x%08X: FatalCode:0x%08X\n",
    (UINTN) RomResult,
    (UINTN) scratchAreaInfo->DebugCode,
    (UINTN) scratchAreaInfo->FatalCode
    ));

  if (RomResult) {
    Status = EFI_SUCCESS;
  } else {
    Status = EFI_SECURITY_VIOLATION;
  }

  return Status;
}
