/*
 * Copyright(c) 2013 Intel Corporation. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * * Neither the name of Intel Corporation nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!	\file CodeSignEmulator.c
 * 	\brief Adds the Hashing and Cryptographic data to a file formatted by the CSSInput tool 
 * 
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <openssl/bn.h>
#include <openssl/crypto.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/objects.h>
#include <openssl/pem.h>
#include <openssl/rand.h>
#include <openssl/rsa.h>
#include <openssl/sha.h>

#define RSA_2048_KEY_SIZE_BITS                2048
#define RSA_2048_KEY_SIZE_BYTES               256
#define RSA_2048_MODULUS_SIZE_BYTES           256
#define RSA_2048_PRIVATE_EXPONENT_SIZE_BYTES  256
#define RSA_2048_PUBLIC_EXPONENT_SIZE_BYTES   4
#define RSA_2048_SIGNATURE_SIZE_BYTES         256
#define SHA_256_DIGEST_SIZE_BYTES             32
#define SHA_MODE                SEC_FW_HASH_SHA256


typedef struct ClantonSecurityHeader
{
	uint32_t CSH_Identifier;					// Unique identifier to sanity check signed module presence
    uint32_t Version;							// Currently unused for Clanton. Maintaining for future use
    uint32_t ModuleSizeBytes;					// Size of the entire module including the Clanton Security Header, RSA Key, RSA Signature and Body
    uint32_t SecurityVersionNumberIndex;		// Index of the SVN to be used for module validation.
    uint32_t SecurityVersionNumber;				// Used to prevent against roll back of modules
    uint32_t RsvdModuleID;						// Currently unused for Clanton. Maintaining for future use
    uint32_t RsvdModuleVendor;					// Vendor Identifier. For Intel products value is 0x00008086
    uint32_t RsvdDate;								// BCD representation of build date as yyyymmdd, where yyyy=4 digit year, mm=1-12, dd=1-31
    uint32_t ModuleHeaderSizeBytes;				// Total length of the header including signed portion, RSA Key, Signature andany padding
    uint32_t HashAlgorithm;						// Hashing Algorithm used during the signing process
    uint32_t CryptoAlgorithm;					// Signing Algorithm used during signing process
    uint32_t KeySizeBytes;       				// Total length of the key data including and padding optionally added by signing tool
    uint32_t SignatureSizeBytes;        		// total length of the signature including any padding optionally added by signing tool
    uint32_t RsvdNextHeaderPointer;
    uint32_t Reserved[8/sizeof(uint32_t)];		// Bytes reserved for future use.
} ClantonSecurityHeader;

typedef struct RSAPublicKey
{
	uint32_t ModulusSizeBytes;
	uint32_t ExponentSizeBytes;
    uint32_t modulus[RSA_2048_MODULUS_SIZE_BYTES/4];
    uint32_t exponent;
} RSAPublicKey;

typedef struct CSSSigningFormat
{
    ClantonSecurityHeader Csh;
    RSAPublicKey SignatureVerificationKey;
    uint32_t Signature[RSA_2048_SIGNATURE_SIZE_BYTES/4];
    char* Body;
} CSSSigningFormat;

typedef enum sec_fw_hash_select_t
{
    SEC_FW_HASH_SHA1 = 0,
    SEC_FW_HASH_SHA256 = 1,
    SEC_FW_HASH_MD5 = 2,
} sec_fw_hash_select_t;

uint32_t ReadBinaryFile(const char* filename, unsigned char** output_buffer, int* output_buffer_size_bytes);
uint32_t GenerateAndVerifySignature(uint32_t* css_header,
                           uint32_t* body, uint32_t body_length_bytes,
                           RSA* key, sec_fw_hash_select_t hash_type,
                           uint32_t* signature_output);
void PrintData(char* string, void* dataInput, uint32_t num_bytes);
void sec_fw_memcpy(uint32_t* dest, uint32_t dest_byte_offset,
                   uint32_t* source, uint32_t source_byte_offset,
                   uint32_t num_bytes, uint32_t dword_endian_swap,
                   uint32_t reverse_memory_order);
void usage();

static const char rnd_seed[] = "string to make the random number generator think it has entropy";




static unsigned char defaultpublicKeyExponent[4] = {0x00,0x00,0x00,0x11};
static unsigned char defaultpublicKeyModulus[256] =
{
0xe5,0x83,0x06,0x33,0x6f,0xda,0xbc,0xcd,0x71,0x17,0xdd,0x5f,0x18,0x91,0x46,0x25,
0x51,0x06,0x75,0xc5,0xd1,0xb2,0x05,0xc5,0x21,0xe3,0x4c,0xf0,0xa0,0xd3,0x16,0x2b,
0xb4,0x7c,0x5d,0x50,0x08,0xf9,0x04,0x85,0xb8,0xf9,0x96,0x68,0xc4,0x11,0x47,0x46,
0x63,0xb7,0x3d,0x84,0x07,0xb9,0x2d,0x12,0xdd,0x98,0x87,0x34,0x15,0x1b,0xbd,0xdc,
0xed,0xfa,0xec,0xe7,0x7a,0xc2,0x2d,0xa7,0xad,0xd6,0x8e,0x6f,0xea,0x6d,0x8f,0x28,
0x22,0x0a,0x53,0x80,0x10,0x6a,0xae,0xab,0xf4,0x62,0xee,0x29,0x16,0x80,0x36,0xcf,
0x37,0xc9,0x39,0x25,0x7b,0xae,0xbe,0x82,0xf2,0x23,0xbe,0x9e,0xbb,0xe9,0x3f,0xeb,
0x3a,0x0e,0xbb,0xa5,0x7b,0xbf,0xf1,0x61,0x88,0x2c,0x36,0xe7,0x63,0x38,0xc6,0x3a,
0x83,0xc0,0x9b,0xf0,0xd9,0xde,0xb5,0xa9,0x33,0x4b,0xda,0xdb,0xa2,0x69,0x1e,0xb4,
0x68,0x9e,0x1b,0x7a,0xe1,0xda,0xa8,0x45,0x26,0x48,0xd6,0xc3,0xe1,0x1a,0xaf,0x9b,
0x11,0x72,0x7a,0x8b,0x88,0x0f,0xd0,0x45,0xae,0x9d,0x63,0x5c,0x96,0x07,0xbd,0xe0,
0x06,0x8c,0x28,0xd6,0x9b,0xf4,0xd7,0x1a,0xa6,0xd5,0x55,0xa0,0xb4,0x28,0x6f,0x6f,
0x77,0x89,0xb4,0x64,0x34,0x1c,0xb7,0x6f,0x1e,0x60,0x0a,0x5a,0x8e,0xb1,0x66,0x2e,
0x5f,0xd6,0x5d,0x35,0x65,0xda,0xc6,0xdc,0x02,0xa6,0xb7,0x1a,0xa8,0xc6,0x49,0xb0,
0x96,0x1a,0x4f,0xba,0xd9,0xff,0x11,0x01,0xaf,0x62,0x8c,0x22,0xb8,0x16,0x8c,0xb8,
0xc5,0x74,0xe9,0x8d,0xee,0xd1,0x0f,0x28,0x3b,0x85,0x83,0x20,0xf4,0x38,0xaa,0xb9
};

static unsigned char defaultprivateKeyExponent[256] =
{
0x36,0x00,0xb6,0x2a,0x38,0x6f,0xb3,0xf4,0x1a,0x9c,0x34,0x16,0x60,0x22,0x2e,0x9f,
0x5e,0x5b,0xdf,0x79,0xd6,0xfc,0xb6,0x10,0x44,0x35,0x7b,0x83,0xe9,0x9b,0x14,0x46,
0x84,0xd1,0xf7,0xd6,0x98,0xb3,0x10,0x1f,0x76,0xd1,0x50,0x91,0x1f,0x13,0x1f,0xd4,
0x53,0xb2,0xa5,0x10,0x01,0xd1,0x37,0xc8,0x34,0x23,0xe3,0x93,0xc8,0xbb,0x3b,0xbb,
0x83,0x4a,0x19,0x9f,0xe0,0xa6,0x28,0xdc,0x28,0xe7,0x30,0x92,0xcd,0xbf,0x6c,0xfa,
0x62,0x5c,0xc8,0x5a,0x5e,0x37,0x38,0x28,0x75,0xbc,0xec,0xbe,0x5f,0xa5,0xb2,0x8b,
0x1c,0x2f,0x58,0xbd,0x86,0x83,0x78,0x1e,0xcf,0x8f,0xf0,0x9d,0xd1,0xdc,0x87,0x82,
0xa4,0x3f,0xb3,0xae,0x77,0x78,0x75,0x07,0xe3,0xce,0x2b,0x09,0x44,0x85,0xd4,0x49,
0x8e,0xcf,0x90,0xcb,0x33,0x48,0xb3,0x1c,0x65,0x33,0x5a,0x49,0x72,0xfd,0xe0,0x10,
0x6c,0x86,0x73,0x27,0x2a,0x28,0x52,0x6b,0xf3,0x2c,0x81,0x4d,0x80,0xf8,0x44,0xe5,
0x55,0x8d,0xb3,0x1b,0x69,0x2f,0x3c,0x5f,0x6a,0x1b,0xc5,0x62,0x1e,0x9b,0x78,0x21,
0xc7,0x85,0x84,0xd3,0x86,0x66,0x1a,0x85,0x42,0xff,0x67,0xc5,0x52,0x3d,0xc4,0x00,
0xc1,0xbd,0xf0,0xfa,0x5b,0x7d,0xc1,0xc7,0x67,0x30,0x26,0xa0,0x84,0x94,0x9f,0xbe,
0xde,0xe1,0x66,0x8a,0x49,0xc4,0x18,0xdf,0x11,0x78,0x6d,0x51,0x03,0x51,0x2b,0x3f,
0xa5,0x9b,0x4a,0xfc,0xbc,0x0c,0x70,0x7d,0x46,0x8e,0xb2,0x72,0x18,0x65,0x2f,0xfa,
0x63,0x1d,0x19,0xba,0x67,0xfd,0x83,0x82,0x2a,0x6b,0x84,0xc6,0x2c,0x03,0x0a,0xf1
};

int main(int argc, char *argv[])
{
    unsigned char publicKeyModulus[RSA_2048_MODULUS_SIZE_BYTES] = {0};
    unsigned char privateKeyExponent[RSA_2048_PRIVATE_EXPONENT_SIZE_BYTES] = {0};
    unsigned char publicKeyExponent[RSA_2048_PUBLIC_EXPONENT_SIZE_BYTES] = {0};
    unsigned char FormattedpublicKeyExponent[RSA_2048_PUBLIC_EXPONENT_SIZE_BYTES] = {0};

    char badarg = 0;
    char* inputFilename = NULL;
    char* outputFilename = NULL;
    char* KeyFileName = NULL;
    int err=0;
    int inputSizeBytes = 0;
    RSA *signatureKey;
    signatureKey = RSA_new();
    unsigned char* inputData = NULL;
    unsigned int useDefaultKey = 1;

    FILE* keyfile_ptr = NULL;

    argc--;
    argv++;

    CRYPTO_malloc_debug_init();
    CRYPTO_dbg_set_options(V_CRYPTO_MDEBUG_ALL);
    CRYPTO_mem_ctrl(CRYPTO_MEM_CHECK_ON);

    RAND_seed(rnd_seed, sizeof rnd_seed); /* or OAEP may fail */

    printf("Collecting Arguments\n");
    while(argc >= 1)
    {
        if (!strcmp(*argv,"-input"))
        {
            argc -= 1;
            if (argc < 1)
            {
                printf("Not enough arguments\n");
                badarg = 1;
            }
            else
            {
                argv++;
                badarg = 0;
                printf("Collecting input filename %s\n", *(argv));
                inputFilename = *(argv);
            }
        }
        else if (!strcmp(*argv,"-output"))
        {
            argc -= 1;
            if (argc < 1)
            {
                printf("Not enough arguments\n");
                badarg = 1;
            }
            else
            {
            	argv++;
                badarg = 0;
                printf("Request for %s output file\n", *(argv));
                outputFilename = *(argv);
            }
        }
        else if (!strcmp(*argv,"-keyfile"))
        {
            argc -= 1;
            if (argc < 1)
            {
                printf("Not enough arguments\n");
                badarg = 1;
            }
            else
            {
              argv++;
                badarg = 0;
                printf("Request for %s key file\n", *(argv));
                KeyFileName = *(argv);
            }
        }
        else
        {
            printf("Unrecognized option  %s\n",*argv);
            badarg = 1;
        }

        if(badarg)
        {
            break;
        }

        argc--;
        argv++;
    }
    printf("Done Collecting Arguments\n");
    if (badarg == 0x0)
    {
        // Check validity of parameters now
        printf("Done Screening Arguments, checking files\n");

        if (inputFilename != NULL)
        {
          if (!ReadBinaryFile(inputFilename, &inputData, &inputSizeBytes))
          {
            badarg = 1;
          }
        }
        else
        {
            printf("Input filename invalid\n");
            badarg = 0x1;
        }
        
        if(KeyFileName != NULL)
        {
	    
	    if((keyfile_ptr = fopen(KeyFileName, "r")) != NULL)
            {
                useDefaultKey = 0;
            }
            else
            {
                badarg = 1;
            }
        }
    }
        
    if (badarg == 0x0 && inputData)
    {
        //int i;
        FILE* outputfile_ptr = NULL;
        CSSSigningFormat* cssSigningFormat = (CSSSigningFormat*)inputData;
        unsigned char* bodyData = &inputData[cssSigningFormat->Csh.ModuleHeaderSizeBytes];

        if (useDefaultKey == 0)
        {
          // read into RSA struct and parse out modulus, private and public exponent
	  printf("Key file: %s specified\n", KeyFileName);
          if( NULL != (signatureKey = PEM_read_RSAPrivateKey(keyfile_ptr, NULL, 0, NULL)))
	  {
	  	printf("RSA Key Check Result: %d\n", RSA_check_key(signatureKey));
          	BN_bn2bin(signatureKey->n, publicKeyModulus);
          	BN_bn2bin(signatureKey->d, privateKeyExponent);
          	BN_bn2bin(signatureKey->e, publicKeyExponent);
                FormattedpublicKeyExponent[0] = publicKeyExponent[3];
          	FormattedpublicKeyExponent[1] = publicKeyExponent[2];
          	FormattedpublicKeyExponent[2] = publicKeyExponent[1];
          	FormattedpublicKeyExponent[3] = publicKeyExponent[0];
	  }
	  else
	  {
	  	fprintf(stderr, "ERROR - KEY FILE INVALID\n");
	  	exit(1);
	  }

          //fclose(keyfile_ptr);
        }
        else
        {
	  printf("No Key File Specified, using default instead\n");
	  signatureKey->n = BN_bin2bn(defaultpublicKeyModulus, sizeof(defaultpublicKeyModulus), signatureKey->n);
    	  signatureKey->e = BN_bin2bn(defaultpublicKeyExponent, sizeof(defaultpublicKeyExponent), signatureKey->e);
    	  signatureKey->d = BN_bin2bn(defaultprivateKeyExponent, sizeof(defaultprivateKeyExponent), signatureKey->d);
	  
	  BN_bn2bin(signatureKey->n, publicKeyModulus);
          BN_bn2bin(signatureKey->d, privateKeyExponent);
          BN_bn2bin(signatureKey->e, publicKeyExponent);
          
          FormattedpublicKeyExponent[0] = publicKeyExponent[3];
          FormattedpublicKeyExponent[1] = publicKeyExponent[2];
          FormattedpublicKeyExponent[2] = publicKeyExponent[1];
          FormattedpublicKeyExponent[3] = publicKeyExponent[0];
          
        }


        if (outputFilename != NULL
                &&  (outputfile_ptr=fopen(outputFilename,"w")) != NULL)
        {
            uint32_t bodySizeBytes = cssSigningFormat->Csh.ModuleSizeBytes - cssSigningFormat->Csh.ModuleHeaderSizeBytes;
            uint32_t* input_signature;
            input_signature = malloc(RSA_2048_SIGNATURE_SIZE_BYTES);
            if (input_signature)
            {
                memset(input_signature, 0x0, RSA_2048_SIGNATURE_SIZE_BYTES);
                printf("Body Size in Bytes = %lu\n", (unsigned long)bodySizeBytes);
                if (!GenerateAndVerifySignature((uint32_t*)&(cssSigningFormat->Csh.CSH_Identifier),
                           (uint32_t*)bodyData, bodySizeBytes,
                           signatureKey, SHA_MODE,
                           input_signature))
                {
                    printf("Problem generating signature\n");
                }

                printf("input Signing complete\n");
                
                cssSigningFormat->SignatureVerificationKey.ModulusSizeBytes = RSA_2048_MODULUS_SIZE_BYTES;
                cssSigningFormat->SignatureVerificationKey.ExponentSizeBytes = RSA_2048_PUBLIC_EXPONENT_SIZE_BYTES;
                
                // Copy the public signature key into the right location
                sec_fw_memcpy(cssSigningFormat->SignatureVerificationKey.modulus, 0x0, (uint32_t*)publicKeyModulus, 0x0, sizeof(publicKeyModulus), 0x0, 0x0);
                sec_fw_memcpy((uint32_t*)&(cssSigningFormat->SignatureVerificationKey.exponent), 0x0, (uint32_t*)FormattedpublicKeyExponent, 0x0, sizeof(publicKeyExponent), 0x0, 0x0);

                // Copy the signature into the right location
                sec_fw_memcpy(cssSigningFormat->Signature, 0x0, input_signature, 0x0, RSA_2048_SIGNATURE_SIZE_BYTES, 0x0, 0x0);

                // Now generate the output file
                printf("Writing binary output file %s\n", outputFilename);
                fwrite(inputData, sizeof(char), inputSizeBytes, outputfile_ptr);
                
            }

            fclose(outputfile_ptr);
            if (input_signature)
            {
                free(input_signature);
                input_signature = NULL;
            }
        }
        else
        {
            printf("Output file invalid\n");
            badarg = 1;
        }

    }
    else
    {
        usage();
    }

    RSA_free(signatureKey);

    CRYPTO_cleanup_all_ex_data();
    ERR_remove_state(0);

    CRYPTO_mem_leaks_fp(stderr);

    if(keyfile_ptr) fclose(keyfile_ptr);
    
    
#ifdef OPENSSL_SYS_NETWARE
    if (err) printf("ERROR: %d\n", err);
#endif
    if (inputData)  free(inputData);
    return err;
}

typedef unsigned char* (*hash_function_ptr)(const unsigned char *d, size_t n, unsigned char *md);

uint32_t GenerateAndVerifySignature(uint32_t* css_header,
                           uint32_t* body, uint32_t body_length_bytes,
                           RSA* key, sec_fw_hash_select_t hash_type,
                           uint32_t* signature_output)
{
    hash_function_ptr hash_function;
    unsigned char* temp_buffer = NULL;
    unsigned char* hash_output = NULL;
    unsigned int md_output_size_bytes = 0x0;
   //unsigned int temp_signature_size;
    int sha_type;
    int pass = 1;
    hash_output = malloc(SHA_256_DIGEST_SIZE_BYTES);
	unsigned char EM[256];
	int index = 0;

    switch (hash_type)
    {
    case SEC_FW_HASH_SHA1:
        {
            hash_function = SHA1;
            md_output_size_bytes = 20;
            sha_type = NID_sha1;
            break;
        }
    case SEC_FW_HASH_SHA256:
        {
            hash_function = SHA256;
            md_output_size_bytes = SHA_256_DIGEST_SIZE_BYTES;
            sha_type = NID_sha256;
            break;
        }
    default:
        {
            printf("Hash type not supported yet\n");
            hash_function = SHA256;
            sha_type = NID_sha256;
            pass = 0;
        }
    }
    if (pass)
    {
        temp_buffer = calloc(body_length_bytes + sizeof(ClantonSecurityHeader), 1);
        printf("Calloc called on %lu bytes\n", (body_length_bytes + sizeof(ClantonSecurityHeader)));
        if ( temp_buffer && hash_output)
        {
            sec_fw_memcpy((uint32_t*)temp_buffer, 0x0, css_header, 0x0, sizeof(ClantonSecurityHeader), 0x0, 0x0);
            sec_fw_memcpy((uint32_t*)temp_buffer, sizeof(ClantonSecurityHeader), body, 0x0, body_length_bytes, 0x0, 0x0);
            printf("Calling hash function with %lu bytes of input_data:\n", (unsigned long)(body_length_bytes+0x40));

            //PrintData("Hash Buffer", temp_buffer, 64+body_length_bytes);
            for(index = 0; index < (body_length_bytes + sizeof(ClantonSecurityHeader)); index++)
            {
            	printf("0x%x, ", temp_buffer[index]);
            }
            printf("\n");

            hash_function(temp_buffer, sizeof(ClantonSecurityHeader) + body_length_bytes, hash_output);

            PrintData("SHA256 digest of input data", hash_output, SHA_256_DIGEST_SIZE_BYTES);

            
	RSA_padding_add_PKCS1_PSS(key, EM, hash_output,  EVP_sha256(), -2);
	RSA_private_encrypt(256, EM, (unsigned char *)signature_output, key, RSA_NO_PADDING);
	//RSA_sign(sha_type, hash_output, md_output_size_bytes, (unsigned char*)signature_output, &temp_signature_size, key);
            printf("Printing Signature Output in hex\n");

            PrintData("RSA Signature",signature_output, 256);

#if 0	// RSA_verify is not compatible with PKCS1_PSS padding.

            printf("Performing signature verification\n");

            if (1 != RSA_verify(sha_type, hash_output, md_output_size_bytes, (unsigned char*)signature_output, temp_signature_size, key))
                printf("RSA signature verification failed\n");
            else
                printf("RSA signature verification passed\n");

            printf("Signature Verification Complete\n");
#endif
        }
    }
    
    if(temp_buffer)
        free(temp_buffer);
            
    if(hash_output)
        free(hash_output);
    
    return pass;
}



void usage()
{
    fprintf(stderr, "Usage: codesignemulator [options]\n");
    fprintf(stderr, "-input   [asciihex|bin] file               input file\n");
    fprintf(stderr, "-output  [asciihex|bin] file               output file\n");
    fprintf(stderr, "(optional)\n");
    fprintf(stderr, "-keyfile [pem] file                        key file\n");
}


uint32_t ReadBinaryFile(const char* filename, unsigned char** output_buffer, int* output_buffer_size_bytes)
{
    uint32_t fileGood = 1;
    int actualFileSize = 0x0;
    FILE* inputFile    = fopen(filename, "rb");
    if (inputFile != NULL)
    {
        fseek(inputFile, 0, SEEK_END);
        actualFileSize = ftell(inputFile);
        fseek(inputFile, 0, SEEK_SET);

        printf("Binary filesize = %d bytes\n", actualFileSize);
        if (actualFileSize > 0)
        {
            *output_buffer = (unsigned char*)malloc(actualFileSize*sizeof(unsigned char));
            if(*output_buffer != NULL)
            {
                fread(*output_buffer, actualFileSize, sizeof(unsigned char), inputFile);
            }
        }
        fclose(inputFile);
    }
    else
        fileGood = 0;
    *output_buffer_size_bytes = actualFileSize;
    return fileGood;
}


void sec_fw_memcpy(uint32_t* dest, uint32_t dest_byte_offset,
                   uint32_t* source, uint32_t source_byte_offset,
                   uint32_t num_bytes, uint32_t dword_endian_swap,
                   uint32_t reverse_memory_order)
{
    uint32_t i;
    unsigned char * temp_dest = (unsigned char*)dest;
    unsigned char * temp_source = (unsigned char*)source;
    for (i=0; i<num_bytes; i++)
    {
      uint32_t dest_index = (dest_byte_offset + i);
      uint32_t source_index = (source_byte_offset + i);
      unsigned char temp_byte = temp_source[source_index];
      temp_dest[dest_index] = temp_byte;
    }
}
void PrintData(char* string, void* dataInput, uint32_t num_bytes)
{
    uint32_t i;
    unsigned char* data_ptr = (unsigned char*)dataInput;
    printf("%s: ", string);
    for (i=0; i<num_bytes; i++)
    {
        printf("%02lx", (unsigned long)data_ptr[i]);
    }
    printf("\n\n");
}
